#define WIN32_NO_STATUS
#include <Windows.h>
#undef WIN32_NO_STATUS
#include <winternl.h>
#include <ntstatus.h>
#include <sddl.h>

#include <cstdio>

#pragma comment(lib, "ntdll")

typedef struct _RTL_RELATIVE_NAME {
  UNICODE_STRING RelativeName;
  HANDLE         ContainingDirectory;
  void* CurDirRef;
} RTL_RELATIVE_NAME, * PRTL_RELATIVE_NAME;

extern "C" {

NTSTATUS NTAPI NtLoadKeyEx(
  POBJECT_ATTRIBUTES   DestinationKeyName,
  POBJECT_ATTRIBUTES   HiveFileName,
  ULONG                Flags,
  HANDLE               TrustClassKey,
  PHANDLE              EventHandle,
  ACCESS_MASK          DesiredAccess,
  PHANDLE              KeyHandle,
  ULONG                Reserved
);

BOOLEAN NTAPI RtlDosPathNameToRelativeNtPathName_U(
  PCWSTR DosFileName,
  PUNICODE_STRING NtFileName,
  PWSTR* FilePath,
  PRTL_RELATIVE_NAME RelativeName
);

NTSTATUS NTAPI RtlInitUnicodeStringEx(
  PUNICODE_STRING Destination,
  PCWSTR          Source
);

}  // extern "C"

BOOL LoadHiveUnderPath(LPCWSTR HivePath,
                       LPCWSTR RegistryPath,
                       ULONG Flags,
                       HKEY htrustkey,
                       PHKEY hkey) {

  UNICODE_STRING HivePathString;
  NTSTATUS Status = RtlInitUnicodeStringEx(&HivePathString, HivePath);
  if (!NT_SUCCESS(Status)) {
    printf("RtlInitUnicodeStringEx failed with error %x\n", Status);
    return FALSE;
  }

  UNICODE_STRING NtHivePath;
  RTL_RELATIVE_NAME RelativeName;
  if (!RtlDosPathNameToRelativeNtPathName_U(HivePathString.Buffer, &NtHivePath, NULL, &RelativeName)) {
    printf("RtlDosPathNameToRelativeNtPathName_U failed\n");
    return FALSE;
  }

  UNICODE_STRING RegistryPathString;
  RtlInitUnicodeString(&RegistryPathString, RegistryPath);

  OBJECT_ATTRIBUTES DestinationKeyName, HiveFileName;
  InitializeObjectAttributes(&DestinationKeyName, &RegistryPathString, 0, NULL, NULL);
  InitializeObjectAttributes(&HiveFileName, &NtHivePath, 0, NULL, NULL);

  Status = NtLoadKeyEx(&DestinationKeyName,
                       &HiveFileName,
                       Flags,
                       htrustkey,
                       NULL,
                       GENERIC_ALL,
                       (PHANDLE)hkey,
                       0);

  if (!NT_SUCCESS(Status)) {
    printf("NtLoadKeyEx failed with error %x\n", Status);
    return FALSE;
  }

  return TRUE;
}

int wmain(int argc, wchar_t** argv) {
  if (argc != 2) {
    wprintf(L"Usage: %s <hive file>\n", argv[0]);
    return 1;
  }

  //
  // Load the app hive with a key inside HKCU as the trust class key.
  //

  HKEY hkcu;
  LONG st = RegOpenKeyExW(HKEY_CURRENT_USER, L"Software", 0, KEY_ALL_ACCESS, &hkcu);

  if (st != ERROR_SUCCESS) {
    printf("RegOpenKeyExA failed with error %d\n", st);
    return 1;
  }

  HKEY happkey;
  if (!LoadHiveUnderPath(argv[1], L"Test", REG_APP_HIVE, hkcu, &happkey)) {
    return 1;
  }

  return 0;
}
